

'''
Rom - the Redis object mapper for Python

Copyright 2013-2016 Josiah Carlson

Released under the LGPL license version 2.1 and version 3 (you can choose
which you'd like to be bound under).
'''

from collections import defaultdict
import json
import warnings

from redis import client
import six

from .columns import (Column, Text, PrimaryKey, ManyToOne, OneToOne, OneToMany,
    MODELS, MODELS_REFERENCED, _on_delete, SKIP_ON_DELETE)
from .exceptions import (ORMError, UniqueKeyViolation, InvalidOperation,
    QueryError, ColumnError, InvalidColumnValue, DataRaceError,
    EntityDeletedError)
from .index import GeneralIndex, GeoIndex
from .query import Query, NUMERIC_TYPES
from .util import (ClassProperty, _connect, session,
    _prefix_score, _script_load, _encode_unique_constraint,
    STRING_SORT_KEYGENS)

_skip = None
_skip = set(globals()) - set(['__doc__'])

_STRING_SORT_KEYGENS = [ss.__name__ for ss in STRING_SORT_KEYGENS]

class _ModelMetaclass(type):
    def __new__(cls, name, bases, dict):
        ns = dict.pop('_namespace', None)
        if ns and not isinstance(ns, six.string_types):
            raise ORMError("The _namespace attribute must be a string, not %s"%type(ns))
        dict['_namespace'] = ns or name
        if name in MODELS or dict['_namespace'] in MODELS:
            raise ORMError("Cannot have two models with the same name (%s) or namespace (%s)"%(name, dict['_namespace']))
        dict['_required'] = required = set()
        dict['_index'] = index = set()
        dict['_unique'] = unique = set()
        dict['_cunique'] = cunique = set()
        dict['_prefix'] = prefix = set()
        dict['_suffix'] = suffix = set()
        dict['_geo'] = geo = {}

        dict['_columns'] = columns = {}
        pkey = None

        # load all columns from any base classes to allow for validation
        odict = {}
        for ocls in reversed(bases):
            if hasattr(ocls, '_columns'):
                odict.update(ocls._columns)
        odict.update(dict)
        dict = odict

        if not any(isinstance(col, PrimaryKey) for col in dict.values()):
            if 'id' in dict:
                raise ColumnError("Cannot have non-primary key named 'id' when no explicit PrimaryKey() is defined")
            dict['id'] = PrimaryKey()

        composite_unique = []
        many_to_one = defaultdict(list)

        # validate all of our columns to ensure that they fulfill our
        # expectations
        for attr, col in dict.items():
            if isinstance(col, Column):
                columns[attr] = col
                if col._required:
                    required.add(attr)
                if col._index:
                    index.add(attr)
                if col._prefix:
                    prefix.add(attr)
                if col._suffix:
                    suffix.add(attr)
                if col._unique:
                    unique.add(attr)

            if isinstance(col, PrimaryKey):
                if pkey:
                    raise ColumnError("Only one primary key column allowed, you have: %s %s"%(
                        pkey, attr)
                    )
                pkey = attr

            if isinstance(col, OneToMany) and not col._column and col._ftable in MODELS:
                # Check to make sure that the foreign ManyToOne/OneToMany table
                # doesn't have multiple references to this table to require an
                # explicit foreign column.
                refs = []
                for _a, _c in MODELS[col._ftable]._columns.items():
                    if isinstance(_c, (ManyToOne, OneToOne)) and _c._ftable == name:
                        refs.append(_a)
                if len(refs) > 1:
                    raise ColumnError("Missing required column argument to OneToMany definition on column %s"%(attr,))

            if isinstance(col, (ManyToOne, OneToOne)):
                many_to_one[col._ftable].append((attr, col))
                MODELS_REFERENCED.setdefault(col._ftable, []).append((dict['_namespace'], attr, col._on_delete))

            if attr == 'unique_together':
                composite_unique = col

            if attr == 'geo_index':
                if not isinstance(col, list) or not all(isinstance(v, GeoIndex) for v in col):
                    raise ORMError("geo_index attribute must be a list of Geoindex() definitions if present")
                for g in col:
                    if g.name in geo:
                        raise ColumnError("Geo index named %r already defined"%(g.name,))
                    geo[g.name] = g

        # verify reverse OneToMany attributes for these ManyToOne/OneToOne
        # attributes if created after referenced models
        for t, cols in many_to_one.items():
            if len(cols) == 1:
                continue
            if t not in MODELS:
                continue
            for _a, _c in MODELS[t]._columns.items():
                if isinstance(_c, OneToMany) and _c._ftable == name and not _c._column:
                    raise ColumnError("Foreign model OneToMany attribute %s.%s missing column argument"%(t, _a))

        # handle multi-column uniqueness constraints
        if composite_unique and isinstance(composite_unique[0], six.string_types):
            composite_unique = [composite_unique]

        seen = {}
        for comp in composite_unique:
            key = tuple(sorted(set(comp)))
            if len(key) == 1:
                raise ColumnError("Single-column unique constraint: %r should be defined via 'unique=True' on the %r column"%(
                    comp, key[0]))
            if key in seen:
                raise ColumnError("Multi-column unique constraint: %r not different than earlier constrant: %r"%(
                    comp, seen[key]))
            for col in key:
                if col not in columns:
                    raise ColumnError("Multi-column unique index %r references non-existant column %r"%(
                        comp, col))
            seen[key] = comp
            cunique.add(key)

        dict['_pkey'] = pkey
        dict['_gindex'] = GeneralIndex(dict['_namespace'])

        MODELS[dict['_namespace']] = MODELS[name] = model = type.__new__(cls, name, bases, dict)
        return model

class AttrDict(dict):
    def __getattr__(self, attr):
        return self.get(attr)
    def __setattr__(self, attr, value):
        self[attr] = value
    def __delattr__(self, attr):
        self.pop(attr, None)

class Model(six.with_metaclass(_ModelMetaclass, object)):
    '''
    This is the base class for all models. You subclass from this base Model
    in order to create a model with columns. As an example::

        class User(Model):
            email_address = String(required=True, unique=True)
            salt = String(default='')
            hash = String(default='')
            created_at = Float(default=time.time, index=True)

    Which can then be used like::

        user = User(email_addrss='user@domain.com')
        user.save() # session.commit() or session.flush() works too
        user = User.get_by(email_address='user@domain.com')
        user = User.get(5)
        users = User.get([2, 6, 1, 7])

    To perform arbitrary queries on entities involving the indices that you
    defined (by passing ``index=True`` on column creation), you access the
    ``.query`` class property on the model::

        query = User.query
        query = query.filter(created_at=(time.time()-86400, time.time()))
        users = query.execute()

    .. note:: You can perform single or chained queries against any/all columns
      that were defined with ``index=True``.

    **Composite/multi-column unique constraints**

    As of version 0.28.0 and later, rom supports the ability for you to have a
    unique constraint involving multiple columns. Individual columns can be
    defined unique by passing the 'unique=True' specifier during column
    definition as always.

    The attribute ``unique_together`` defines those groups of columns that when
    taken together must be unique for ``.save()`` to complete successfully.
    This will work almost exactly the same as Django's ``unique_together``, and
    is comparable to SQLAlchemy's ``UniqueConstraint()``.

    Usage::

        class UniquePosition(Model):
            x = Integer()
            y = Integer()

            unique_together = [
                ('x', 'y'),
            ]

    .. note:: If one or more of the column values on an entity that is part of a
        unique constrant is None in Python, the unique constraint won't apply.
        This is the typical behavior of nulls in unique constraints inside both
        MySQL and Postgres.
    '''
    def __init__(self, **kwargs):
        self._new = not kwargs.pop('_loading', False)
        model = self._namespace
        self._data = {}
        self._last = {}
        self._modified = False
        self._deleted = False
        self._init = False
        for attr in self._columns:
            cval = kwargs.get(attr, None)
            data = (model, attr, cval, not self._new)
            if self._new and attr == self._pkey and cval:
                raise InvalidColumnValue("Cannot pass primary key on object creation")
            setattr(self, attr, data)
            if cval != None:
                if not isinstance(cval, six.string_types):
                    cval = self._columns[attr]._to_redis(cval)
                self._last[attr] = cval
        self._init = True
        # note: this is a lie, don't use it outside of query.py
        if kwargs.pop('_bypass_session_entirely', False):
            session.add(self)

    def _before_insert(self):
        "Called before a new entity is saved to Redis"

    def _before_update(self):
        "Called before an entity is saved to Redis"

    def _before_delete(self):
        "Called before an entity is deleted"

    def _after_insert(self):
        "Called after a new entity has been saved to Redis"

    def _after_update(self):
        "Called after a previously-saved entity has been saved to Redis"

    def _after_delete(self):
        "Called after a previously-deleted entity has been deleted from Redis"

    @ClassProperty
    def _connection(cls):
        return _connect(cls)

    def refresh(self, force=False):
        if self._deleted:
            return
        if self._modified and not force:
            raise InvalidOperation("Cannot refresh a modified entity without passing force=True to override modified data")
        if self._new:
            raise InvalidOperation("Cannot refresh a new entity")

        conn = _connect(self)
        data = conn.hgetall(self._pk)
        if six.PY3:
            data = dict((k.decode(), v.decode()) for k, v in data.items())
        self.__init__(_loading=True, **data)

    @property
    def _pk(self):
        return '%s:%s'%(self._namespace, getattr(self, self._pkey))

    @classmethod
    def _apply_changes(cls, old, new, full=False, delete=False, is_new=False):
        conn = _connect(cls)
        pk = old.get(cls._pkey) or new.get(cls._pkey)
        if not pk:
            raise ColumnError("Missing primary key value")

        model = cls._namespace
        columns = cls._columns
        changes = 0
        keys = set()
        scores = {}
        data = {}
        unique = {}
        deleted = []
        udeleted = {}
        prefix = []
        suffix = []
        geo = []
        redis_data = {}

        # update individual columns
        for attr in cls._columns:
            ikey = None
            if attr in cls._unique:
                ikey = "%s:%s:uidx"%(model, attr)

            ca = columns[attr]
            roval = old.get(attr)
            oval = ca._from_redis(roval) if roval is not None else None

            nval = new.get(attr)
            rnval = ca._to_redis(nval) if nval is not None else None
            if rnval is not None:
                redis_data[attr] = rnval

            # Add/update standard index
            if ca._keygen and not delete and nval is not None and (ca._index or ca._prefix or ca._suffix):
                generated = ca._keygen(attr, new)
                if not generated:
                    # No index entries, we'll clean out old entries later
                    pass

                elif isinstance(generated, (list, tuple, set)):
                    if ca._index:
                        for k in generated:
                            keys.add('%s:%s'%(attr, k))

                    if ca._prefix:
                        for k in generated:
                            prefix.append([attr, k])

                    if ca._suffix:
                        for k in generated:
                            if six.PY2 and isinstance(k, str) and isinstance(ca, Text):
                                try:
                                    suffix.append([attr, k.decode('utf-8')[::-1].encode('utf-8')])
                                except UnicodeDecodeError:
                                    suffix.append([attr, k[::-1]])
                            else:
                                suffix.append([attr, k[::-1]])

                elif isinstance(generated, dict):
                    if ca._index:
                        for k, v in generated.items():
                            if not k:
                                scores[attr] = v
                            elif v in (None, ''):
                                # mixed index type support
                                keys.add('%s:%s'%(attr, k))
                            else:
                                scores['%s:%s'%(attr, k)] = v

                    if ca._prefix:
                        if ca._keygen.__name__ not in _STRING_SORT_KEYGENS:
                            warnings.warn("Prefix indexes are currently not enabled for non-standard keygen functions", stacklevel=2)
                        else:
                            prefix.append([attr, nval if ca._keygen.__name__ in ('SIMPLE', 'IDENTITY') else nval.lower()])

                    if ca._suffix:
                        if ca._keygen.__name__ not in _STRING_SORT_KEYGENS:
                            warnings.warn("Prefix indexes are currently not enabled for non-standard keygen functions", stacklevel=2)
                        else:
                            ex = (lambda x:x) if ca._keygen.__name__ in ('SIMPLE', 'IDENTITY') else (lambda x:x.lower())
                            if six.PY2 and isinstance(nval, str) and isinstance(ca, Text):
                                try:
                                    suffix.append([attr, ex(nval.decode('utf-8')[::-1]).encode('utf-8')])
                                except UnicodeDecodeError:
                                    suffix.append([attr, ex(nval[::-1])])
                            else:
                                suffix.append([attr, ex(nval[::-1])])

                else:
                    raise ColumnError("Don't know how to turn %r into a sequence of keys"%(generated,))

            if nval == oval and not full:
                continue

            changes += 1

            # Delete removed columns
            if nval is None and oval is not None:
                deleted.append(attr)
                if ikey:
                    udeleted[attr] = roval
                continue

            # Add/update column value
            if nval is not None:
                data[attr] = rnval

            # Add/update unique index
            if ikey:
                if six.PY2 and not isinstance(roval, str):
                    roval = columns[attr]._to_redis(roval)
                if oval is not None and roval != rnval:
                    udeleted[attr] = oval
                if rnval is not None:
                    unique[attr] = rnval

        # Add/update multi-column unique constraint
        for uniq in cls._cunique:
            attr = ':'.join(uniq)

            odata = [old.get(c) for c in uniq]
            ndata = [new.get(c) for c in uniq]
            ndata = [columns[c]._to_redis(nv) if nv is not None else None for c, nv in zip(uniq, ndata)]

            if odata != ndata and None not in odata:
                udeleted[attr] = _encode_unique_constraint(odata)

            if None not in ndata:
                unique[attr] = _encode_unique_constraint(ndata)

        for name in cls._geo:
            idx = cls._geo[name]
            val = idx.callback(AttrDict(new))
            if isinstance(val, dict):
                val = [val]
            for v in val:
                if 'lon' in v and 'lat' in v:
                    geo.append((name, v['lon'], v['lat']))
                else:
                    raise ORMError("Lon/Lat pair for geo index is not a dictionary of {'lon': ..., 'lat': ...}")

        id_only = str(pk)
        old_data = [] if is_new else ([(cls._pkey, str(pk))] + [(k, old.get(k)) for k in data if k in old])
        redis_writer_lua(conn, cls._pkey, model, id_only, unique, udeleted,
            deleted, data, list(keys), scores, prefix, suffix, geo, old_data,
            delete)

        return changes, redis_data

    def to_dict(self):
        '''
        Returns a copy of all data assigned to columns in this entity. Useful
        for returning items to JSON-enabled APIs. If you want to copy an
        entity, you should look at the ``.copy()`` method.
        '''
        return dict(self._data)

    def save(self, full=False, force=False):
        '''
        Saves the current entity to Redis. Will only save changed data by
        default, but you can force a full save by passing ``full=True``.

        If the underlying entity was deleted and you want to re-save the entity,
        you can pass ``force=True`` to force a full re-save of the entity.
        '''
        # handle the pre-commit hooks
        was_new = self._new
        if was_new:
            self._before_insert()
        else:
            self._before_update()

        new = self.to_dict()
        ret, data = self._apply_changes(
            self._last, new, full or self._new or force, is_new=self._new or force)
        self._last = data
        self._new = False
        self._modified = False
        self._deleted = False
        # handle the post-commit hooks
        if was_new:
            self._after_insert()
        else:
            self._after_update()
        return ret

    def delete(self, **kwargs):
        '''
        Deletes the entity immediately. Also performs any on_delete operations
        specified as part of column definitions.
        '''
        if kwargs.get('skip_on_delete_i_really_mean_it') is not SKIP_ON_DELETE:
            # handle the pre-commit hook
            self._before_delete()
            # handle any foreign key references + cascade options
            _on_delete(self)

        session.forget(self)
        self._apply_changes(self._last, {}, delete=True)
        self._modified = True
        self._deleted = True
        # handle the post-commit hooks
        if kwargs.get('skip_on_delete_i_really_mean_it') is not SKIP_ON_DELETE:
            self._after_delete()

    def copy(self):
        '''
        Creates a shallow copy of the given entity (any entities that can be
        retrieved from a OneToMany relationship will not be copied).
        '''
        x = self.to_dict()
        x.pop(self._pkey)
        return self.__class__(**x)

    @classmethod
    def get(cls, ids):
        '''
        Will fetch one or more entities of this type from the session or
        Redis.

        Used like::

            MyModel.get(5)
            MyModel.get([1, 6, 2, 4])

        Passing a list or a tuple will return multiple entities, in the same
        order that the ids were passed.
        '''
        conn = _connect(cls)
        # prepare the ids
        single = not isinstance(ids, (list, tuple, set, frozenset))
        if single:
            ids = [ids]
        pks = ['%s:%s'%(cls._namespace, id) for id in map(int, ids)]
        # get from the session, if possible
        out = list(map(session.get, pks))
        # if we couldn't get an instance from the session, load from Redis
        if None in out:
            pipe = conn.pipeline(True)
            idxs = []
            # Fetch missing data
            for i, data in enumerate(out):
                if data is None:
                    idxs.append(i)
                    pipe.hgetall(pks[i])
            # Update output list
            for i, data in zip(idxs, pipe.execute()):
                if data:
                    if six.PY3:
                        data = dict((k.decode(), v.decode()) for k, v in data.items())
                    out[i] = cls(_loading=True, **data)
            # Get rid of missing models
            out = [x for x in out if x]
        if single:
            return out[0] if out else None
        return out

    @classmethod
    def get_by(cls, **kwargs):
        '''
        This method offers a simple query method for fetching entities of this
        type via attribute numeric ranges (such columns must be ``indexed``),
        or via ``unique`` columns.

        Some examples::

            user = User.get_by(email_address='user@domain.com')
            # gets up to 25 users created in the last 24 hours
            users = User.get_by(
                created_at=(time.time()-86400, time.time()),
                _limit=(0, 25))

        Optional keyword-only arguments:

            * *_limit* - A 2-tuple of (offset, count) that can be used to
              paginate or otherwise limit results returned by a numeric range
              query
            * *_numeric* - An optional boolean defaulting to False that forces
              the use of a numeric index for ``.get_by(col=val)`` queries even
              when ``col`` has an existing unique index

        If you would like to make queries against multiple columns or with
        multiple criteria, look into the Model.query class property.

        .. note:: rom will attempt to use a unique index first, then a numeric
            index if there was no unique index. You can explicitly tell rom to
            only use the numeric index by using ``.get_by(..., _numeric=True)``.
        .. note:: Ranged queries with `get_by(col=(start, end))` will only work
            with columns that use a numeric index.
        '''
        conn = _connect(cls)
        model = cls._namespace
        # handle limits and query requirements
        _limit = kwargs.pop('_limit', ())
        if _limit and len(_limit) != 2:
            raise QueryError("Limit must include both 'offset' and 'count' parameters")
        elif _limit and not all(isinstance(x, six.integer_types) for x in _limit):
            raise QueryError("Limit arguments must both be integers")
        if len(kwargs) != 1:
            raise QueryError("We can only fetch object(s) by exactly one attribute, you provided %s"%(len(kwargs),))

        _numeric = bool(kwargs.pop('_numeric', None))

        for attr, value in kwargs.items():
            plain_attr = attr.partition(':')[0]
            if isinstance(value, tuple) and len(value) != 2:
                raise QueryError("Range queries must include exactly two endpoints")

            # handle unique index lookups
            if attr in cls._unique and (plain_attr not in cls._index or not _numeric):
                if isinstance(value, tuple):
                    raise QueryError("Cannot query a unique index with a range of values")
                single = not isinstance(value, list)
                if single:
                    value = [value]
                qvalues = list(map(cls._columns[attr]._to_redis, value))
                ids = [x for x in conn.hmget('%s:%s:uidx'%(model, attr), qvalues) if x]
                if not ids:
                    return None if single else []
                return cls.get(ids[0] if single else ids)

            if plain_attr not in cls._index:
                raise QueryError("Cannot query on a column without an index")

            if isinstance(value, NUMERIC_TYPES) and not isinstance(value, bool):
                value = (value, value)

            if isinstance(value, tuple):
                # this is a numeric range query, we'll just pull it directly
                args = list(value)
                for i, a in enumerate(args):
                    # Handle the ranges where None is -inf on the left and inf
                    # on the right when used in the context of a range tuple.
                    args[i] = ('-inf', 'inf')[i] if a is None else cls._columns[attr]._to_redis(a)
                if _limit:
                    args.extend(_limit)
                ids = conn.zrangebyscore('%s:%s:idx'%(model, attr), *args)
                if not ids:
                    return []
                return cls.get(ids)

            # defer other index lookups to the query object
            query = cls.query.filter(**{attr: value})
            if _limit:
                query = query.limit(*_limit)
            return query.all()

    @ClassProperty
    def query(cls):
        '''
        Returns a ``Query`` object that refers to this model to handle
        subsequent filtering.
        '''
        return Query(cls)

_redis_writer_lua = _script_load('''
local namespace = ARGV[1]
local id = ARGV[2]
local row_key = string.format('%s:%s', namespace, id)
local is_delete = cjson.decode(ARGV[12])

-- [1] string.format("%s", d) will truncate d to the first null value, so we
--     can't rely on string.format() where we can reasonably expect nulls.

if not is_delete then
    -- check to make sure we don't have a data race condition
    local updated = {}
    for i, pair in ipairs(cjson.decode(ARGV[13])) do
        local odata = redis.call('HGET', row_key, pair[1])
        if odata ~= pair[2] then
            table.insert(updated, pair[1])
        end
    end
    if #updated > 0 then
        return cjson.encode({race=updated})
    end
end

-- check and update unique column constraints
for i, write in ipairs({false, true}) do
    for col, value in pairs(cjson.decode(ARGV[3])) do
        local key = string.format('%s:%s:uidx', namespace, col)
        if write then
            redis.call('HSET', key, value, id)
        else
            local known = redis.call('HGET', key, value)
            if known ~= id and known ~= false then
                return cjson.encode({unique=col})
            end
        end
    end
end

-- remove deleted unique constraints
for col, value in pairs(cjson.decode(ARGV[4])) do
    local key = string.format('%s:%s:uidx', namespace, col)
    local known = redis.call('HGET', key, value)
    if known == id then
        redis.call('HDEL', key, value)
    end
end

-- remove deleted columns
local deleted = cjson.decode(ARGV[5])
if #deleted > 0 then
    redis.call('HDEL', string.format('%s:%s', namespace, id), unpack(deleted))
end

-- update changed/added columns
local data = cjson.decode(ARGV[6])
if #data > 0 then
    redis.call('HMSET', row_key, unpack(data))
end

-- remove old index data, update util.clean_index_lua when changed
local idata = redis.call('HGET', namespace .. '::', id)
local _changes = 0
if idata then
    idata = cjson.decode(idata)
    while #idata < 5 do
        idata[#idata + 1] = {}
    end
    for i, key in ipairs(idata[1]) do
        redis.call('SREM', string.format('%s:%s:idx', namespace, key), id)
        -- see note [1]
        redis.call('SREM', namespace .. ':' .. key .. ':idx', id)
        _changes = _changes + 1
    end
    for i, key in ipairs(idata[2]) do
        redis.call('ZREM', string.format('%s:%s:idx', namespace, key), id)
        -- see note [1]
        redis.call('ZREM', namespace .. ':' .. key .. ':idx', id)
        _changes = _changes + 1
    end
    for i, data in ipairs(idata[3]) do
        local key = string.format('%s:%s:pre', namespace, data[1])
        local mem = string.format('%s\0%s', data[2], id)
        redis.call('ZREM', key, mem)
        -- see note [1]
        local key = namespace .. ':' .. data[1] .. ':pre'
        local mem = data[2] .. '\0' .. id
        redis.call('ZREM', key, mem)
        _changes = _changes + 1
    end
    for i, data in ipairs(idata[4]) do
        if data[1] and data[2] then
            local key = string.format('%s:%s:suf', namespace, data[1])
            local mem = string.format('%s\0%s', data[2], id)
            redis.call('ZREM', key, mem)
            -- see note [1]
            local key = namespace .. ':' .. data[1] .. ':suf'
            local mem = data[2] .. '\0' .. id
            redis.call('ZREM', key, mem)
            _changes = _changes + 1
        end
    end
    for i, data in ipairs(idata[5]) do
        local key = namespace .. ':' .. data .. ':geo'
        redis.call('ZREM', key, id)
        _changes = _changes + 1
    end
end

if is_delete then
    redis.call('DEL', string.format('%s:%s', namespace, id))
    redis.call('HDEL', namespace .. '::', id)
    return cjson.encode({changes=_changes})
end

-- add new key index data
local nkeys = cjson.decode(ARGV[7])
for i, key in ipairs(nkeys) do
    redis.call('SADD', namespace .. ':' .. key .. ':idx', id)
end

-- add new scored index data
local nscored = {}
for key, score in pairs(cjson.decode(ARGV[8])) do
    redis.call('ZADD', namespace .. ':' .. key .. ':idx', score, id)
    nscored[#nscored + 1] = key
end

-- add new prefix data
local nprefix = {}
for i, data in ipairs(cjson.decode(ARGV[9])) do
    local key = namespace .. ':' .. data[1] .. ':pre'
    local mem = data[2] .. '\0' .. id
    redis.call('ZADD', key, data[3], mem)
    nprefix[#nprefix + 1] = {data[1], data[2]}
end

-- add new suffix data
local nsuffix = {}
for i, data in ipairs(cjson.decode(ARGV[10])) do
    local key = namespace .. ':' .. data[1] .. ':suf'
    local mem = data[2] .. '\0' .. id
    redis.call('ZADD', key, data[3], mem)
    nsuffix[#nsuffix + 1] = {data[1], data[2]}
end

-- add new geo data
local ngeo = {}
for i, data in ipairs(cjson.decode(ARGV[11])) do
    local key = namespace .. ':' .. data[1] .. ':geo'
    redis.call('GEOADD', key, data[2], data[3], id)
    nsuffix[#nsuffix + 1] = data[1]
end

if not is_delete then
    -- update known index data
    local encoded = cjson.encode({nkeys, nscored, nprefix, nsuffix, ngeo})
    redis.call('HSET', namespace .. '::', id, encoded)
end
return cjson.encode({changes=#nkeys + #nscored + #nprefix + #nsuffix + #ngeo + _changes})
''')

def _fix_bytes(d):
    if six.PY2:
        raise TypeError
    if isinstance(d, bytes):
        return d.decode('latin-1')
    raise TypeError

def redis_writer_lua(conn, pkey, namespace, id, unique, udelete, delete,
                     data, keys, scored, prefix, suffix, geo, old_data, is_delete):
    '''
    ... Actually write data to Redis. This is an internal detail. Please don't
    call me directly.
    '''
    ldata = []
    for pair in data.items():
        ldata.extend(pair)

    for item in prefix:
        item.append(_prefix_score(item[-1]))
    for item in suffix:
        item.append(_prefix_score(item[-1]))

    data = [json.dumps(x, default=_fix_bytes) for x in
            (unique, udelete, delete, ldata, keys, scored, prefix, suffix, geo, is_delete, old_data)]
    result = _redis_writer_lua(conn, [], [namespace, id] + data)

    if isinstance(result, client.BasePipeline):
        # we're in a pipelined write situation, don't parse the pipeline :P
        return

    if six.PY3:
        result = result.decode()

    result = json.loads(result)
    if 'unique' in result:
        result = result['unique']
        raise UniqueKeyViolation(
            "Value %r for %s:%s:uidx not distinct (failed for pk=%s)"%(
                unique[result], namespace, result, id),
            namespace, id)

    if 'race' in result:
        result = result['race']
        if pkey in result:
            raise EntityDeletedError(
                "Entity %s:%s deleted by another writer; use .save(force=True) to re-save"%(
                    namespace, id),
                namespace, id)

        raise DataRaceError(
            "%s:%s Column(s) %r updated by another writer, write aborted!"%(
                namespace, id, result),
            namespace, id)

__all__ = [k for k, v in globals().items() if getattr(v, '__doc__', None) and k not in _skip]




'''
Rom - the Redis object mapper for Python

Copyright 2013-2016 Josiah Carlson

Released under the LGPL license version 2.1 and version 3 (you can choose
which you'd like to be bound under).
'''

from __future__ import print_function
import base64
from collections import namedtuple
from datetime import datetime, timedelta
from decimal import Decimal as _Decimal
import sys
import time
import unittest
import warnings

import redis
import six

from rom import util

util.CONNECTION = redis.Redis(db=15)
connect = util._connect

from rom import *
from rom.exceptions import *

string = String if six.PY2 else Text


def global_setup():
    c = connect(None)
    for p in ('RomTest*', 'RestrictA*', 'RestrictB*'):
        keys = c.keys(p)
        if keys:
            c.delete(*keys)
    from rom.columns import MODELS
    Model = MODELS['Model']
    for k, v in MODELS.copy().items():
        if v is not Model:
            del MODELS[k]

def get_state():
    c = connect(None)
    data = []
    for k in c.keys('*'):
        k = k.decode() if six.PY3 else k
        t = c.type(k)
        t = t.decode() if six.PY3 else t
        if t == 'string':
            data.append((k, c.get(k)))
        elif t == 'list':
            data.append((k, c.lrange(k, 0, -1)))
        elif t == 'set':
            data.append((k, c.smembers(k)))
        elif t == 'hash':
            data.append((k, c.hgetall(k)))
        else:
            data.append((k, c.zrange(k, 0, -1, withscores=True)))
    data.sort()
    return data

_now = datetime.utcnow()
_now_time = time.time()

def _default_time():
    return _now_time

class TestORM(unittest.TestCase):
    def setUp(self):
        session.rollback()

    # for Python 2.6
    def assertIsNone(self, arg):
        assert arg is None
    def assertIs(self, arg1, arg2):
        assert arg1 is arg2

    def test_basic_model(self):
        class RomTestBasicModel(Model):
            val = Integer()
            oval = Integer(default=7)
            created_at = Float(default=_default_time)
            req = Text(required=True)

        self.assertRaises(ColumnError, RomTestBasicModel)
        self.assertRaises(InvalidColumnValue, lambda: RomTestBasicModel(oval='t', req='X'))
        self.assertRaises(MissingColumn, lambda: RomTestBasicModel(created_at=7))

        # try object saving/loading
        x = RomTestBasicModel(val=1, req=b"h\xd0\x80llo".decode('utf-8'))
        x.save()
        id = x.id
        x = x.to_dict()

        y = RomTestBasicModel.get(id)
        yd = y.to_dict()
        ## cax = x.pop('created_at'); cay = yd.pop('created_at')
        self.assertEqual(x, yd)
        ## self.assertTrue(abs(cax - cay) < .005, cax-cay)

        # try object copying
        zd = y.copy().to_dict()
        ## caz = zd.pop('created_at')
        self.assertNotEqual(yd, zd)
        zd.pop('id')
        yd.pop('id')
        self.assertEqual(yd, zd)
        ## self.assertTrue(abs(cay-caz) < .005, cay-caz)

        util.session.rollback()
        self.assertEqual([m.id for m in RomTestBasicModel.query], [id])
        self.assertEqual([m.id for m in RomTestBasicModel.query.iter_result(no_hscan=True)], [id])

    def test_unique_index(self):
        class RomTestIndexModel(Model):
            key = Text(required=True, unique=True)

        self.assertRaises(MissingColumn, RomTestIndexModel)
        item = RomTestIndexModel(key="hello")
        item.save()
        item2 = item.copy()
        self.assertRaises(UniqueKeyViolation, item2.save)

        m = RomTestIndexModel.get_by(key="hello")
        self.assertTrue(m)
        self.assertEqual(m.id, item.id)
        self.assertTrue(m is item)

    def test_unique_integer_index(self):
        class RomTestIndexModel2(Model):
            key = Integer(unique=True)

        item = RomTestIndexModel2(key=5)
        item.save()
        # verify this works when there is no data
        RomTestIndexModel2().save()
        RomTestIndexModel2().save()

        m = RomTestIndexModel2.get_by(key=5)
        self.assertTrue(m)
        self.assertEqual(m.id, item.id)
        self.assertTrue(m is item)

    def test_foreign_key(self):
        def foo():
            class RomTestBFkey1(Model):
                bad = ManyToOne("RomTestBad", 'no action')
            RomTestBFkey1()
        self.assertRaises(ORMError, foo)

        def foo2():
            class RomTestBFkey2(Model):
                bad = OneToMany("RomTestBad")
            RomTestBFkey2()
        self.assertRaises(ORMError, foo2)

        class RomTestFkey1(Model):
            fkey2 = ManyToOne("RomTestFkey2", 'no action')
        class RomTestFkey2(Model):
            fkey1 = OneToMany("RomTestFkey1")

        x = RomTestFkey2()
        y = RomTestFkey1(fkey2=x) # implicitly saves x
        y.save()

        xid = x.id
        yid = y.id
        x = y = None
        y = RomTestFkey1.get(yid)
        self.assertEqual(y.fkey2.id, xid)
        fk1 = y.fkey2.fkey1

        self.assertEqual(len(fk1), 1)
        self.assertEqual(fk1[0].id, y.id)

    def test_foreign_key_delete(self):
        class RomTestM2Ofk(Model):
            ref = ManyToOne("RomTestO2Mfk", 'no action')
        class RomTestO2Mfk(Model):
            lst = OneToMany("RomTestM2Ofk")

        x = RomTestO2Mfk()
        x.save()
        ys = [RomTestM2Ofk(ref=x) for z in range(5)]
        [y.save() for y in ys]

        del ys[0].ref
        ys[0].save()
        self.assertEqual(ys[0].ref, None)
        ys[0].ref = x
        ys[0].save()
        self.assertEqual(ys[1].ref, x)

        ys[1].ref = None
        ys[1].save()
        self.assertEqual(ys[1].ref, None)
        self.assertEqual(len(x.lst), 4)
        # Test for double-delete
        ys[1].ref = None
        ys[1].save()
        ys[1].ref = x
        ys[1].save()
        self.assertEqual(ys[1].ref, x)

    def test_unique(self):
        class RomTestUnique(Model):
            attr = Text(unique=True)

        a = RomTestUnique(attr='hello')
        b = RomTestUnique(attr='hello2')
        a.save()
        b.save()
        b.attr = 'hello'
        self.assertRaises(UniqueKeyViolation, b.save)

        c = RomTestUnique(attr='hello')
        self.assertRaises(UniqueKeyViolation, c.save)

        a.delete()
        b.save()

    def test_saving(self):
        class RomTestNormal(Model):
            attr = Text()

        self.assertTrue(RomTestNormal().save())
        self.assertTrue(RomTestNormal(attr='hello').save())
        x = RomTestNormal()
        self.assertTrue(x.save())
        self.assertFalse(x.save())
        session.commit()

        self.assertTrue(x is RomTestNormal.get(x.id))

    def test_index(self):
        class RomTestIndexedModel(Model):
            attr = Text(index=True, keygen=FULL_TEXT)
            attr2 = Text(index=True, keygen=FULL_TEXT)
            attr3 = Integer(index=True)
            attr4 = Float(index=True)
            attr5 = Decimal(index=True)

        x = RomTestIndexedModel(
            attr='hello world',
            attr2='how are you doing?',
            attr3=7,
            attr4=4.5,
            attr5=_Decimal('2.643'),
        )
        x.save()
        RomTestIndexedModel(
            attr='world',
            attr3=100,
            attr4=-1000,
            attr5=_Decimal('2.643'),
        ).save()

        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello').count(), 1)
        self.assertEqual(RomTestIndexedModel.query.filter(attr2='how').filter(attr2='are').count(), 1)
        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello').filter(attr2='how').filter(attr2='are').count(), 1)
        self.assertRaises(QueryError, lambda: RomTestIndexedModel.query.filter(attr='hello', noattr='bad'))
        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello', attr3=(None, None)).count(), 1)
        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello', attr3=(None, 10)).count(), 1)
        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello', attr3=(None, 10)).execute()[0].id, 1)
        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello', attr3=(5, None)).count(), 1)
        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello', attr3=(5, 10), attr4=(4,5), attr5=(2.5, 2.7)).count(), 1)
        first = RomTestIndexedModel.query.filter(attr='hello', attr3=(5, 10), attr4=(4,5), attr5=(2.5, 2.7)).first()
        self.assertTrue(first)
        self.assertTrue(first is x)
        self.assertEqual(RomTestIndexedModel.query.filter(attr='hello', attr3=(10, 20), attr4=(4,5), attr5=(2.5, 2.7)).count(), 0)
        self.assertEqual(RomTestIndexedModel.query.filter(attr3=100).count(), 1)
        self.assertEqual(RomTestIndexedModel.query.filter(attr='world', attr5=_Decimal('2.643')).count(), 2)

        results = RomTestIndexedModel.query.filter(attr='world').order_by('attr4').execute()
        self.assertEqual([y.id for y in results], [2,1])

        for i in range(50):
            RomTestIndexedModel(attr3=i)
        session.commit()
        session.rollback()

        self.assertEqual(len(RomTestIndexedModel.get_by(attr3=(10, 25))), 16)
        self.assertEqual(len(RomTestIndexedModel.get_by(attr3=(10, 25), _limit=(0,5))), 5)

        key = RomTestIndexedModel.query.filter(attr='hello').filter(attr2='how').filter(attr2='are').cached_result(30)
        conn = connect(None)
        self.assertTrue(conn.ttl(key) <= 30)
        self.assertEqual(conn.zcard(key), 1)
        conn.delete(key)
        self.assertRaises(QueryError, lambda: RomTestIndexedModel.query.order_by('attr6'))

        # Only the first call raises the warning, so only bother to call it
        # for our first pass through tests (non-Lua case).
        with warnings.catch_warnings(record=True) as w:
            RomTestIndexedModel.query.order_by('attr')
            self.assertEqual(len(w), 1)

    def test_alternate_models(self):
        ctr = [0]
        class RomTestAlternate(object):
            def __init__(self, id=None):
                if id is None:
                    id = ctr[0]
                    ctr[0] += 1
                self.id = id

            @classmethod
            def get(self, id):
                return RomTestAlternate(id)

        class RomTestFModel(Model):
            attr = ForeignModel(RomTestAlternate)

        a = RomTestAlternate()
        ai = a.id
        i = RomTestFModel(attr=a).id
        session.commit()   # two lines of magic to destroy session history
        session.rollback() #
        del a

        f = RomTestFModel.get(i)
        self.assertEqual(f.attr.id, ai)

    def test_model_connection(self):
        class RomTestFoo(Model):
            pass

        class RomTestBar(Model):
            _conn = redis.Redis(db=14)

        RomTestBar._conn.delete('RomTestBar:id:')

        x = RomTestFoo()
        x.save()
        y = RomTestBar()
        y.save()
        self.assertEqual(RomTestFoo._connection, util.CONNECTION)
        self.assertEqual(x._connection, util.CONNECTION)
        self.assertNotEqual(RomTestBar._connection, util.CONNECTION)
        self.assertNotEqual(y._connection, util.CONNECTION)

        self.assertEqual(RomTestBar._conn.get('RomTestBar:id:').decode(), '1')
        self.assertEqual(util.CONNECTION.get('RomTestBar:id:'), None)
        RomTestBar.get(1).delete()
        RomTestBar._conn.delete('RomTestBar:id:')
        k = RomTestBar._conn.keys('RomTest*')
        if k:
            RomTestBar._conn.delete(*k)

    def test_entity_caching(self):
        class RomTestGoo(Model):
            pass

        f = RomTestGoo()
        i = f.id
        session.commit()

        for j in range(10):
            RomTestGoo()

        g = RomTestGoo.get(i)
        self.assertIs(f, g)

    def test_index_preservation(self):
        """ Edits to unrelated columns should not remove the index of other
        columns. Issue: https://github.com/josiahcarlson/rom/issues/2. """

        class RomTestM(Model):
            u = Text(unique=True)
            i = Integer(index=True)
            unrelated = Text()

        RomTestM(u='foo', i=11).save()

        m = RomTestM.get_by(u='foo')
        m.unrelated = 'foobar'
        self.assertEqual(len(RomTestM.get_by(i=11)), 1)
        m.save()
        self.assertEqual(len(RomTestM.get_by(i=11)), 1)
        self.assertEqual(len(RomTestM.get_by(i=(10, 12))), 1)

    def test_json_multisave(self):
        class RomTestJsonTest(Model):
            col = Json()

        d = {'hello': 'world'}
        x = RomTestJsonTest(col=d)
        x.save()
        del x
        for i in range(5):
            x = RomTestJsonTest.get(1)
            self.assertEqual(x.col, d)
            x.save(full=True)
            session.rollback()

    def test_boolean(self):
        class RomTestBooleanTest(Model):
            col = Boolean(index=True)

        RomTestBooleanTest(col=True).save()
        RomTestBooleanTest(col=1).save()
        RomTestBooleanTest(col=False).save()
        RomTestBooleanTest(col='').save()
        RomTestBooleanTest(col=None).save() # None is considered "not data", so is ignored
        y = RomTestBooleanTest()
        yid = y.id
        y.save()
        del y
        self.assertEqual(len(RomTestBooleanTest.get_by(col=True)), 2)
        self.assertEqual(len(RomTestBooleanTest.get_by(col=False)), 2)
        session.rollback()
        x = RomTestBooleanTest.get(1)
        x.col = False
        x.save()
        self.assertEqual(len(RomTestBooleanTest.get_by(col=True)), 1)
        self.assertEqual(len(RomTestBooleanTest.get_by(col=False)), 3)
        self.assertEqual(len(RomTestBooleanTest.get_by(col=True)), 1)
        self.assertEqual(len(RomTestBooleanTest.get_by(col=False)), 3)
        y = RomTestBooleanTest.get(yid)
        self.assertEqual(y.col, None)

    def test_datetimes(self):
        class RomTestDateTimesTest(Model):
            col1 = DateTime(index=True)
            col2 = Date(index=True)
            col3 = Time(index=True)

        dtt = RomTestDateTimesTest(col1=_now, col2=_now.date(), col3=_now.time())
        dtt.save()
        session.commit()
        del dtt
        self.assertEqual(len(RomTestDateTimesTest.get_by(col1=_now)), 1)
        self.assertEqual(len(RomTestDateTimesTest.get_by(col2=_now.date())), 1)
        self.assertEqual(len(RomTestDateTimesTest.get_by(col3=_now.time())), 1)

    def test_deletion(self):
        class RomTestDeletionTest(Model):
            col1 = Text(index=True, keygen=FULL_TEXT)

        x = RomTestDeletionTest(col1="this is a test string that should be indexed")
        session.commit()
        self.assertEqual(len(RomTestDeletionTest.get_by(col1='this')), 1)

        x.delete()
        self.assertEqual(len(RomTestDeletionTest.get_by(col1='this')), 0)

        session.commit()
        self.assertEqual(len(RomTestDeletionTest.get_by(col1='this')), 0)

    def test_empty_query(self):
        class RomTestEmptyQuery(Model):
            col1 = Text()

        x = RomTestEmptyQuery()
        x.save()
        session.forget(x)
        self.assertEqual(len(RomTestEmptyQuery.query.all()), 1)
        self.assertEqual(RomTestEmptyQuery.query.first().id, x.id)
        self.assertEqual(RomTestEmptyQuery.query.count(), 1)
        self.assertEqual(RomTestEmptyQuery.query.limit(0, 10).count(), 1)
        self.assertEqual(RomTestEmptyQuery.query.limit(1, 10).count(), 0)

    def test_refresh(self):
        class RomTestRefresh(Model):
            col = Text()

        d = RomTestRefresh(col='hello')
        d.save()
        d.col = 'world'
        self.assertRaises(InvalidOperation, d.refresh)
        d.refresh(True)
        self.assertEqual(d.col, 'hello')
        d.col = 'world'
        session.refresh(d, force=True)
        self.assertEqual(d.col, 'hello')
        d.col = 'world'
        session.refresh_all(force=True)
        self.assertEqual(d.col, 'hello')
        self.assertRaises(InvalidOperation, RomTestRefresh(col='boo').refresh)

    def test_datetime(self):
        class RomTestDT(Model):
            created_at = DateTime(default=datetime.utcnow)
            event_datetime = DateTime(index=True)

        x = RomTestDT()
        x.event_datetime = datetime.utcnow()
        x.save()
        RomTestDT(event_datetime=datetime.utcnow()).save()
        session.rollback() # clearing the local cache

        self.assertEqual(RomTestDT.get_by(event_datetime=(datetime(2000, 1, 1), datetime(2000, 1, 1))), [])
        self.assertEqual(len(RomTestDT.get_by(event_datetime=(datetime(2000, 1, 1), datetime.utcnow()))), 2)

    def test_prefix_suffix_pattern(self):
        class RomTestPSP(Model):
            col = Text(prefix=True, suffix=True, keygen=FULL_TEXT)
            col2 = Text(prefix=True, suffix=True, keygen=SIMPLE)
            col3 = Text(prefix=True, suffix=True, keygen=SIMPLE_CI)

        x = RomTestPSP(col="hello world how are you doing, join us today",
                       col2="This is just another Test",
                       col3="And This is yet Another")
        x.save()

        self.assertEqual(RomTestPSP.query.startswith(col='he').count(), 1)
        self.assertEqual(RomTestPSP.query.startswith(col='notthere').count(), 0)
        self.assertEqual(RomTestPSP.query.endswith(col='rld').count(), 1)
        self.assertEqual(RomTestPSP.query.endswith(col='bad').count(), 0)
        self.assertEqual(RomTestPSP.query.like(col='?oin?').count(), 1)
        self.assertEqual(RomTestPSP.query.like(col='*oin+').count(), 1)
        self.assertEqual(RomTestPSP.query.like(col='oin').count(), 0)
        self.assertEqual(RomTestPSP.query.like(col='+oin').like(col='wor!d').count(), 1)
        self.assertEqual(RomTestPSP.query.like(col='*1').count(), 0)
        self.assertEqual(RomTestPSP.query.like(col='*r').count(), 1)

        self.assertEqual(RomTestPSP.query.startswith(col2="This is just").count(), 1)
        self.assertEqual(RomTestPSP.query.startswith(col2="this is just").count(), 0)
        self.assertEqual(RomTestPSP.query.endswith(col2="another Test").count(), 1)
        self.assertEqual(RomTestPSP.query.endswith(col2="another test").count(), 0)
        self.assertEqual(RomTestPSP.query.like(col2="This?is").count(), 1)
        self.assertEqual(RomTestPSP.query.like(col2="this?is").count(), 0)

        self.assertEqual(RomTestPSP.query.startswith(col3="and this is").count(), 1)
        self.assertEqual(RomTestPSP.query.startswith(col3="And This Is").count(), 1)
        self.assertEqual(RomTestPSP.query.startswith(col3="And This isn't").count(), 0)
        self.assertEqual(RomTestPSP.query.endswith(col3="yet another").count(), 1)
        self.assertEqual(RomTestPSP.query.endswith(col3="Yet another").count(), 1)
        self.assertEqual(RomTestPSP.query.endswith(col3="and another").count(), 0)
        self.assertEqual(RomTestPSP.query.like(col3="And?This").count(), 1)
        self.assertEqual(RomTestPSP.query.like(col3="and?this").count(), 1)
        self.assertEqual(RomTestPSP.query.like(col3="*this is*").count(), 1)
        self.assertEqual(RomTestPSP.query.like(col3="nope").count(), 0)

        ## self.assertEqual(RomTestPSP.query.like(col="world\0").count(), 1)
        ## self.assertEqual(RomTestPSP.query.startswith(col="world").count(), 1)

    def test_unicode_text(self):
        ch = unichr(0xfeff) if six.PY2 else chr(0xfeff)
        pre = ch + 'hello'
        suf = 'hello' + ch

        class RomTestUnicode1(Model):
            col = Text(index=True, unique=True, keygen=FULL_TEXT)
            col2 = Text(index=True, keygen=IDENTITY)

        RomTestUnicode1(col=pre, col2=pre).save()
        RomTestUnicode1(col=suf, col2=suf).save()

        self.assertEqual(RomTestUnicode1.query.filter(col=pre, col2=pre).count(), 1)
        self.assertEqual(RomTestUnicode1.query.filter(col=suf, col2=suf).count(), 1)
        self.assertTrue(RomTestUnicode1.get_by(col=pre))
        self.assertTrue(RomTestUnicode1.get_by(col2=pre))
        self.assertTrue(RomTestUnicode1.get_by(col=suf))
        self.assertTrue(RomTestUnicode1.get_by(col2=suf))

        class RomTestUnicode2(Model):
            col = Text(prefix=True, suffix=True, keygen=FULL_TEXT)
            col2 = Text(prefix=True, suffix=True, keygen=IDENTITY)

        RomTestUnicode2(col=pre, col2=pre).save()
        RomTestUnicode2(col=suf, col2=suf).save()

        self.assertEqual(RomTestUnicode2.query.startswith(col="h").count(), 1)
        self.assertEqual(RomTestUnicode2.query.startswith(col2="h").count(), 1)
        self.assertEqual(RomTestUnicode2.query.startswith(col=ch).count(), 1)
        self.assertEqual(RomTestUnicode2.query.startswith(col2=ch).count(), 1)
        self.assertEqual(RomTestUnicode2.query.endswith(col="o").count(), 1)
        self.assertEqual(RomTestUnicode2.query.endswith(col2="o").count(), 1)
        self.assertEqual(RomTestUnicode2.query.endswith(col=ch).count(), 1)
        self.assertEqual(RomTestUnicode2.query.endswith(col2=ch).count(), 1)

    def test_infinite_ranges(self):
        """ Infinite range lookups via None in tuple.
        The get_by method accepts None as an argument to range based
        lookups.  A range lookup is done by passing a tuple as the value
        to the kwarg representing the column in question such as,
        Model.get_by(some_column=(small, large)).  The left and right side
        are inclusive. """

        class RomTestInfRange(Model):
            dt = DateTime(index=True)
            num = Integer(index=True)

        start_dt = datetime(2000, 1, 1)
        for x in range(3):
            RomTestInfRange(dt=start_dt+timedelta(days=x), num=x)
        session.commit()
        ranges = (
            (dict(dt=(start_dt-timedelta(days=1), None)), 3),
            (dict(dt=(start_dt, None)), 3),
            (dict(dt=(start_dt+timedelta(days=0.5), None)), 2),
            (dict(dt=(start_dt+timedelta(days=1), None)), 2),
            (dict(dt=(start_dt+timedelta(days=1.5), None)), 1),
            (dict(dt=(start_dt+timedelta(days=2), None)), 1),
            (dict(dt=(start_dt+timedelta(days=2.5), None)), 0),

            (dict(dt=(None, start_dt+timedelta(days=2.5))), 3),
            (dict(dt=(None, start_dt+timedelta(days=2))), 3),
            (dict(dt=(None, start_dt+timedelta(days=1.5))), 2),
            (dict(dt=(None, start_dt+timedelta(days=1))), 2),
            (dict(dt=(None, start_dt+timedelta(days=0.5))), 1),
            (dict(dt=(None, start_dt)), 1),
            (dict(dt=(None, start_dt-timedelta(days=1))), 0),

            (dict(num=(-1, None)), 3),
            (dict(num=(0, None)), 3),
            (dict(num=(0.5, None)), 2),
            (dict(num=(1, None)), 2),
            (dict(num=(1.5, None)), 1),
            (dict(num=(2, None)), 1),
            (dict(num=(2.5, None)), 0),

            (dict(num=(None, 2.5)), 3),
            (dict(num=(None, 2)), 3),
            (dict(num=(None, 1.5)), 2),
            (dict(num=(None, 1)), 2),
            (dict(num=(None, 0.5)), 1),
            (dict(num=(None, 0)), 1),
            (dict(num=(None, -1)), 0),
        )
        for i, (kwargs, count) in enumerate(ranges):
            try:
                st = 1
                self.assertEqual(len(RomTestInfRange.get_by(**kwargs)), count)
                st = 2
                self.assertEqual(RomTestInfRange.query.filter(**kwargs).count(), count)
                st = 3
                self.assertEqual(len(RomTestInfRange.query.filter(**kwargs).execute()), count)
            except Exception:
                msg = 'test %d step %d: range %s, expect: %d' % (i, st, kwargs, count)
                print(msg)
                raise

    def test_big_int(self):
        """ Ensure integers that overflow py2k int work. """

        class RomTestBigInt(Model):
            num = Integer()

        numbers = [
            1,
            200,
            1<<15,
            1<<63,
            1<<128,
            1<<256,
        ]

        for i, num in enumerate(numbers):
            RomTestBigInt(num=num).save()
            session.commit()
            session.rollback()
            echo = RomTestBigInt.get(i+1).num
            self.assertEqual(num, echo)

    def test_cleanup(self):
        """ Ensure no side effects are left in the db after a delete. """
        redis = connect(None)

        class RomTestCleanupA(Model):
            foo = Text()
            blist = OneToMany('RomTestCleanupB')

        class RomTestCleanupB(Model):
            bar = Text()
            a = ManyToOne('RomTestCleanupA', 'no action')

        a = RomTestCleanupA(foo='foo')
        a.save()
        b = RomTestCleanupB(bar='foo', a=a)
        b.save()
        b.delete()
        self.assertFalse(redis.hkeys('RomTestCleanupB:%d' % b.id))
        a.delete()
        self.assertFalse(redis.hkeys('RomTestCleanupA:%d' % a.id))

        # Test delete() where a column value does not change. This affects
        # the writer logic which checks for deltas as a means to determine
        # what keys should be removed from the redis hash bucket.
        a = RomTestCleanupA(foo='foo')
        a.save()
        b = RomTestCleanupB(bar='foo', a=a)
        b.save()
        aid = a.id
        apk = a._pk
        self.assertTrue(b.a)

        a.delete()  # Nullify FK on b.
        self.assertFalse(redis.hkeys('RomTestCleanupA:%d' % a.id))
        # verify removal from the session object
        self.assertFalse(RomTestCleanupA.get(aid))
        self.assertFalse(apk in session.known)
        self.assertFalse(apk in session.wknown)
        session.rollback() # XXX purge session cache

        b = RomTestCleanupB.get(b.id)
        self.assertFalse(b.a)

        b.delete()  # Nullify FK on b.
        self.assertFalse(redis.hkeys('RomTestCleanupB:%d' % b.id))

    def test_delete_writethrough(self):
        """ Verify that a Model.delete() writes through backing and session. """

        class RomTestDelete(Model):
            pass

        # write-through backing
        a = RomTestDelete()
        a.save()
        a.delete()
        session.commit()
        session.rollback()
        self.assertIsNone(RomTestDelete.get(a.id))

        # write-through cache auto-commit (session)
        a = RomTestDelete()
        a.save()
        a.delete()
        self.assertIsNone(RomTestDelete.get(a.id))

        # write-through cache force-commit (session)
        a = RomTestDelete()
        a.save()
        a.delete()
        session.commit()
        self.assertIsNone(RomTestDelete.get(a.id))

    def test_restrict_on_delete(self):
        """ Verify that Restrict is thrown when there is a foreign object referencing
            the deleted object."""

        class RomTestCascadeRestrict(Model):
            alist = OneToMany('RomTestRestrictA')

        class RomTestRestrictA(Model):
            foo = Text()
            cref = ManyToOne('RomTestCascadeRestrict', 'cascade')
            blist = OneToMany('RomTestRestrictB')

        class RomTestRestrictB(Model):
            bar = Text()
            a = ManyToOne('RomTestRestrictA', 'restrict')

        c = RomTestCascadeRestrict()
        c.save()
        a = RomTestRestrictA(foo='foo', cref=c)
        a.save()
        b = RomTestRestrictB(bar='foo', a=a)
        b.save()
        self.assertRaises(RestrictError, c.delete)
        self.assertRaises(RestrictError, a.delete)
        del a.cref
        a.save()
        c.delete()
        self.assertRaises(RestrictError, a.delete)

    def test_restrict_on_delete_121(self):
        """ Verify that Restrict is thrown when there is a foreign object referencing
            the deleted object."""

        class RomTestCascadeRestrictOne(Model):
            pass

        class RomTestRestrictAOne(Model):
            foo = Text()
            c = OneToOne('RomTestCascadeRestrictOne', 'cascade')

        class RomTestRestrictBOne(Model):
            a = OneToOne('RomTestRestrictAOne', 'restrict')
            bar = Text()

        c = RomTestCascadeRestrictOne()
        c.save()
        a = RomTestRestrictAOne(foo='foo', c=c)
        a.save()
        b = RomTestRestrictBOne(bar='foo', a=a)
        b.save()

        self.assertRaises(RestrictError, c.delete)
        self.assertRaises(RestrictError, a.delete)
        self.assertEqual(RomTestRestrictBOne.query.filter(a=NOT_NULL).count(), 1)
        del b.a
        b.save()
        c.delete()

    def _test_restrict_skip_on_delete(self):
        # Functionality currently disabled, might be useful in the future, but
        # currently differs from existing Postgres/MySQL/Oracle/MSSql behavior
        class RomTestRoot(Model):
            leaf = OneToMany('RomTestLeaf')
        class RomTestNode(Model):
            root = OneToOne('RomTestRoot', 'cascade')
            node = OneToOne('RomTestNode', 'cascade')
            leaf = OneToMany('RomTestLeaf')
        class RomTestLeaf(Model):
            node = ManyToOne('RomTestNode', 'restrict')
            root = ManyToOne('RomTestRoot', 'cascade')

        root = RomTestRoot()
        root.save()
        node1 = RomTestNode(root=root)
        node1.save()
        node2 = RomTestNode(node=node1) # assignment to an attribute auto-saves
        node2.save()

        l1 = RomTestLeaf(root=root, node=node2)
        l2 = RomTestLeaf(root=root, node=node2)
        l1.save()
        l2.save()

        ids = [
            (RomTestRoot, root.id),
            (RomTestNode, node1.id),
            (RomTestNode, node2.id),
            (RomTestLeaf, l1.id),
            (RomTestLeaf, l2.id),
        ]

        # resulting structure:
        #    root.node -> node1;.node -> node2;.leaf -> [l1, l2]
        #    root.leaf -> [l1, l2]
        # Cascading deletes use a breadth-first traversal of references, so the
        # queue of deletes when starting from the root will look like:
        #    [root, node1, l1, l2, node2]
        # Because l1 and l2 are deleted as part of the cascading delete from
        # the root, and *not* from the restricted delete from node2 (or node1),
        # a delete of the root should cascade fully (node1 is not necessary for
        # this to work; it's there for other tests).
        # Given the above, a delete from node1 or node2 should cause a
        # RestrictError.
        del root, node1, node2, l1, l2
        session.rollback()
        # from node1 gets a queue of:
        # [node1, node2] -restrict-> [l1, l2]
        self.assertRaises(RestrictError, RomTestNode.get(ids[2][1]).delete)
        session.rollback()
        # from node2 gets a queue of:
        # [node2] -restrict-> [l1, l2]
        self.assertRaises(RestrictError, RomTestNode.get(ids[1][1]).delete)

        session.rollback()
        RomTestRoot.get(ids[0][1]).delete()

        # make sure the deletions happened
        for mdl, id in ids:
            self.assertEqual(mdl.get(id), None)

    def test_on_delete_extra(self):
        def fail1():
            class RomTestFail(Model):
                col = OneToOne('OtherModel', 'set default', required=True)
        def fail2():
            class RomTestFail(Model):
                col = OneToOne('OtherModel', 'set null', required=True)
        self.assertRaises(ColumnError, fail1)
        self.assertRaises(ColumnError, fail2)

        class RomTestSetNull(Model):
            col = OneToOne('RomTestSetNull', 'set null')

        a = RomTestSetNull()
        b = RomTestSetNull(col=a)
        b.save()
        ia = a.id
        self.assertEqual(len(b.get_by(col=ia)), 1)
        a.delete()
        self.assertEqual(len(b.get_by(col=ia)), 0)

    def test_prefix_suffix1(self):
        class RomTestPerson(Model):
            name = Text(prefix=True, suffix=True, index=True, keygen=FULL_TEXT)

        names = ['Acasaoi', 'Maria Williamson IV', 'Rodrigo Howe',
            'Mr. Willow Goldner', 'Melody Prohaska', 'Hulda Botsford',
            'Lester Swaniawski MD', 'Vilma Mohr Sr.', 'Pierre Moen',
            'Beau Streich', 'Mrs. Laron Morar III', 'bmliodasas',
            'Jewell Stroman', 'Garfield Stark', 'Dr. Ignatius Kuvalis PhD',
            'Nikita Okuneva', 'Daija Turcotte', 'Royce Halvorson',
            'Tess Schimmel', 'Ms. Monte Heathcote', 'Johann Glover',
            'Kade Lueilwitz', 'bsaasao', 'Casper Pouros',
            'Miss Griffin Corkery II', 'Cierra Volkman V', 'Sean McLaughlin',
            'Cmlio', 'Cdsdmlio', 'Dasao', 'Dioasu', 'Eioasu', 'Fasao',
            'Emilie Towne II', 'G h o', 'Jhorecgssd',
            'H Mrs. Newton Murazik Sr.  Zidfdfoaxaol dfgdfggf ',
            'Zidfdfoasaol dfgdfggf Mrs. Newton Murazik Sr.  ',
            'Perry Ankunding', 'Dusty Kessler', 'Jacinthe Bechtelar',
            'Dr. Jordan Hintz PhD', 'Miss Monty Kuvalis']
        for name in names:
            RomTestPerson(name=name)
        session.commit()

        self.assertEqual(RomTestPerson.query.like(name='*asao*').count(), 5)

    def test_prefix_suffix2(self):
        class RomTestPerson2(Model):
            idPerson = string(prefix=True, suffix=True, index=True, keygen=FULL_TEXT)
            description = string(prefix=True, suffix=True, index=True, keygen=FULL_TEXT)
        data = [
            ["8947589545872", "ayuntamientodeciudad"],
            ["8947589545872", "ayuntamientodeguipuzcoa"],
            ["8947589545872", "ayuntamientodepalencia"],
            ["8947589545872", "ayuntamientodeciudad"],
            ["8937589569872", "ayuntamientodeburgos"],
            ["8947689545872", "ayuntamientodeburgos"],
            ["8947689545872", "ayuntamientodeburgos"],
            ["894789545872", "ayuntamientodeciudad"]
        ]
        cols = ['idPerson', 'description']
        for d in data:
            RomTestPerson2(**dict(zip(cols, d)))
        session.commit()

        self.assertEqual(RomTestPerson2.query.startswith(idPerson='89375').filter(description="ayuntamientodeburgos").count(), 1)
        self.assertEqual(RomTestPerson2.query.like(idPerson='*94*').filter(description="ayuntamientodeburgos").count(), 2)

    def test_null_session(self):
        class RomTestNullSession(Model):
            data = string()

        x = RomTestNullSession(data="test")
        x.save()
        session.rollback()

        util.use_null_session()
        y = RomTestNullSession.get(x.id)
        self.assertNotEqual(x, y)
        self.assertEqual(util.session.get(x._pk), None)

        util.use_rom_session()
        z = RomTestNullSession.get(x.id)
        self.assertEqual(util.session.get(x._pk), z)

        util.session.rollback()
        del x, y, z

    def test_saving_after_modifying_datetime_object(self):
        class RomTestObjectChangeAndSave(Model):
            col1 = DateTime()

        x = RomTestObjectChangeAndSave(col1=_now)
        self.assertTrue(x.save())

        y = RomTestObjectChangeAndSave.get(x.id)
        y.col1 = datetime.now()
        self.assertTrue(y.save())

    def test_index_cleared(self):
        class RomTestIndexClear(Model):
            col1 = Integer(index=True)

        for i in range(10):
            RomTestIndexClear(col1=i)

        session.commit()
        session.rollback()
        self.assertEqual(RomTestIndexClear.query.count(), 10)
        self.assertEqual(len(list(RomTestIndexClear.query)), 10)
        self.assertEqual(len(RomTestIndexClear.query.all()), 10)
        self.assertEqual(len(list(RomTestIndexClear.query.iter_result(no_hscan=True))), 10)
        self.assertEqual(len(list(RomTestIndexClear.query.limit(4, 10))), 6)
        self.assertEqual(len(list(RomTestIndexClear.query.limit(2, 4))), 4)
        self.assertEqual(len(RomTestIndexClear.query.limit(2, 4).all()), 4)

        for j in range(1, 11):
            RomTestIndexClear.get(j).delete()
        conn = connect(None)
        self.assertEqual(conn.hgetall('RomTestIndexClear::'), {})
        self.assertEqual(RomTestIndexClear.query.count(), 0)

    def test_multi_col_unique_index(self):
        class RomTestCompositeUnique(Model):
            col1 = Integer()
            col2 = Integer()
            col3 = string()

            unique_together = [
                ('col1', 'col2', 'col3'),
            ]

        for c1 in range(10):
            for c2 in range(10):
                for c3 in ('a', 'b', 'c', 'test', 'blah'):
                    RomTestCompositeUnique(col1=c1, col2=c2, col3=c3).save()

        self.assertRaises(UniqueKeyViolation, RomTestCompositeUnique(col1=5, col2=5, col3='c').save)

        a = RomTestCompositeUnique.get(5)
        a.col1 = 4
        self.assertRaises(UniqueKeyViolation, a.save)
        a.col3 = 'world'
        a.save()
        a.col1 = 0
        a.col3 = 'blah'
        a.save()

    def test_iter_result(self):
        # also tests alternate column names for the primary key, and an index
        # on the primary key
        class RomTestIterResult(Model):
            _id = PrimaryKey(index=True)
            id = Integer()
            col1 = Integer(index=True)

        for i in range(1, 51):
            RomTestIterResult(col1=i).save()

        session.rollback()
        total = 0
        for it in RomTestIterResult.query.order_by('col1').iter_result(30, 10):
            total += it.col1
        session.rollback()
        self.assertEqual(total, 50 * 51 / 2)

        # test no decoding
        total = 0
        for it in RomTestIterResult.query.select('col1', decode=False).order_by('col1').iter_result(30, 10):
            total += int(it['col1'])
        self.assertEqual(total, 50 * 51 / 2)

        # test with decoding
        total = 0
        for it in RomTestIterResult.query.select('col1', decode=True).order_by('col1').iter_result(30, 10):
            total += it['col1']
        self.assertEqual(total, 50 * 51 / 2)

        from rom.query import _namedtuple_data_factory as ntf
        from rom.query import _tuple_data_factory
        from rom.query import _list_data_factory

        # test alternate factory function output
        total = 0
        for it in RomTestIterResult.query.select('col1', decode=True, ff=ntf).order_by('col1').iter_result(30, 10):
            total += it.col1
        self.assertEqual(total, 50 * 51 / 2)

        # test alternate factory function output
        total = 0
        for it in RomTestIterResult.query.select('col1', decode=True, ff=_tuple_data_factory).order_by('col1').iter_result(30, 10):
            total += it[0]
        self.assertEqual(total, 50 * 51 / 2)

        # test alternate factory function output
        total = 0
        for it in RomTestIterResult.query.select('col1', decode=True, ff=_list_data_factory).order_by('col1').iter_result(30, 10):
            total += it[0]
        self.assertEqual(total, 50 * 51 / 2)

        # test bare column factory function output
        def mf(ignore):
            def make(data):
                return data[0]
            return make

        total = 0
        for it in RomTestIterResult.query.select('col1', include_pk=False, decode=True, ff=mf):
            total += int(it)
        self.assertEqual(total, 50 * 51 / 2)

        # also test open-ended get_by() on the indexed primary key
        self.assertEqual(len(RomTestIterResult.get_by(_id=(None, None))), 50)
        # also test order-by on the indexed primary key
        self.assertEqual(len(RomTestIterResult.query.order_by('_id').all()), 50)
        session.rollback()

        # check select on the plain order by id...
        total = 0
        for it in RomTestIterResult.query.select('_id', decode=True, ff=ntf).order_by('_id'):
            total += it.id
        self.assertEqual(total, 50 * 51 / 2)

        # Let's get all the entities
        session.rollback()
        self.assertEqual(len(RomTestIterResult.query.all()), 50)

        # and get all the entities without an order by...
        total = 0
        for it in RomTestIterResult.query.select('col1', include_pk=False, decode=True, ff=ntf):
            total += it.col1
        self.assertEqual(total, 50 * 51 / 2)

        # And let's verify that it used the primary key index to iterate...
        c = connect(None)
        c.delete('%s:%s:idx'%(RomTestIterResult._namespace, RomTestIterResult._pkey))
        self.assertEqual(len(RomTestIterResult.query.all()), 0)
        session.rollback()

        # hack the primary key column to force using HSCAN when available
        RomTestIterResult._columns['_id']._index = False

        total = 0
        for it in RomTestIterResult.query.select('col1', include_pk=False, ff=ntf):
            total += int(it.col1)
        self.assertEqual(total, 50 * 51 / 2)

        self.assertEqual(len(RomTestIterResult.query.all()), 50)

        # And now use the naive query
        total = 0
        for it in RomTestIterResult.query.select('col1', include_pk=False, ff=ntf).iter_result(no_hscan=True):
            total += int(it.col1)
        self.assertEqual(total, 50 * 51 / 2)

        self.assertEqual(len(list(RomTestIterResult.query.iter_result(no_hscan=True))), 50)

    def test_foreign_model_references(self):
        class RomTestM2O(Model):
            col1 = ManyToOne('RomTestO2M', 'no action')
            col2 = ManyToOne('RomTestO2M', 'no action')

        def r():
            class RomTestO2M(Model):
                col1 = OneToMany('RomTestM2O')

        def ok():
            class RomTestO2M(Model):
                col1 = OneToMany('RomTestM2O', 'col2')

        self.assertRaises(ColumnError, r)
        ok()

        class RomTestO2M_(Model):
            col1 = OneToMany('RomTestM2O_')

        def r():
            class RomTestM2O_(Model):
                col1 = ManyToOne('RomTestO2M_', 'no action')
                col2 = ManyToOne('RomTestO2M_', 'no action')

        self.assertRaises(ColumnError, r)

        class RomTestO2M__(Model):
            col1 = OneToMany('RomTestM2O__', 'col2')

        class RomTestM2O__(Model):
            col1 = ManyToOne('RomTestO2M__', 'no action')
            col2 = ManyToOne('RomTestO2M__', 'no action')

    def test_clean_old_index(self):
        class RomTestCleanOld(Model):
            _namespace = 'RomTestNamespacedCleanup'
            col1 = Integer(index=True)
            col2 = string(index=True, keygen=FULL_TEXT)
            col3 = string(unique=True)

        now = str(time.time())

        a = RomTestCleanOld(col1=6, col2="this is content that should be indexed", col3=now)
        a.save()
        id = a.id
        self.assertEqual(len(RomTestCleanOld.get_by(col1=6)), 1)
        self.assertEqual(len(RomTestCleanOld.get_by(col1=(5, 7))), 1)
        self.assertEqual(len(RomTestCleanOld.get_by(col2='content')), 1)
        self.assertTrue(RomTestCleanOld.get_by(col3=now))
        session.rollback()
        del a
        c = connect(None)
        self.assertEqual(c.hlen('RomTestNamespacedCleanup::'), 1)
        self.assertEqual(RomTestCleanOld.query.count(), 1)
        self.assertEqual(c.scard('RomTestNamespacedCleanup:col2:content:idx'), 1)
        self.assertEqual(c.zcard('RomTestNamespacedCleanup:col1:idx'), 1)
        self.assertEqual(c.hlen('RomTestNamespacedCleanup:col3:uidx'), 1)

        self.assertEqual(c.delete('RomTestNamespacedCleanup:%s'%id), 1)

        with warnings.catch_warnings(record=True) as w:
            all(util.clean_old_index(RomTestCleanOld, force_hscan=None))
            self.assertEqual(len(w), 1)

        self.assertEqual(c.hlen('RomTestNamespacedCleanup::'), 0)
        self.assertEqual(RomTestCleanOld.query.count(), 0)
        self.assertEqual(c.scard('RomTestNamespacedCleanup:col2:content:idx'), 0)
        self.assertEqual(c.zcard('RomTestNamespacedCleanup:col1:idx'), 0)
        # can't clean out unique index when force_hscan is None - aka HSCAN disabled
        self.assertEqual(c.hlen('RomTestNamespacedCleanup:col3:uidx'), 1)
        c.delete('RomTestNamespacedCleanup:col3:uidx')

        # okay, now test for longer scan/clear.
        minid = int(c.get('RomTestNamespacedCleanup:id:')) + 1
        _count = 200
        for i in range(minid, minid+_count):
            RomTestCleanOld(col1=i, col3=str(i)).save()
        session.rollback()

        version = list(map(int, c.info()['redis_version'].split('.')[:2]))
        has_hscan = version >= [2, 8]
        if has_hscan:
            self.assertEqual(len(RomTestCleanOld.query.all()), _count)

        to_delete = list(range(minid, minid + _count, 37))
        c.delete(*['RomTestNamespacedCleanup:%i'%i for i in to_delete])
        self.assertTrue(all(c.hexists('RomTestNamespacedCleanup::', i) for i in to_delete))
        all(util.clean_old_index(RomTestCleanOld, 10, force_hscan=has_hscan))
        self.assertTrue(all(not c.hexists('RomTestNamespacedCleanup::', i) for i in to_delete))
        if has_hscan:
            self.assertTrue(all(not c.hexists('RomTestNamespacedCleanup:col3:uidx', i) for i in to_delete))

        to_delete = list(range(minid+29, minid + _count, 29))
        c.delete(*['RomTestNamespacedCleanup:%i'%i for i in to_delete])
        self.assertTrue(all(c.hexists('RomTestNamespacedCleanup::', i) for i in to_delete))
        # should cause a warning
        with warnings.catch_warnings(record=True) as w:
            all(util.clean_old_index(RomTestCleanOld, 10, force_hscan=None))
            self.assertEqual(len(w), 1)
        self.assertTrue(all(not c.hexists('RomTestNamespacedCleanup::', i) for i in to_delete))
        # We can't really clean out unique indexes when hscan is disabled or not
        # available. :/
        self.assertTrue(all(c.hexists('RomTestNamespacedCleanup:col3:uidx', i) for i in to_delete))

    def test_mutli_query(self):
        class RomTestIndexMultiCol(Model):
            attr1 = string(required=True, index=True, keygen=FULL_TEXT)
            attr2 = string(required=True, index=True, keygen=FULL_TEXT)

        a = RomTestIndexMultiCol(attr1='548ef7ee7b77b93ab41ksjh3', attr2='2')
        a.save()

        self.assertEqual(len(RomTestIndexMultiCol.query.filter(attr1='548ef7ee7b77b93ab41ksjh3').execute()), 1)
        self.assertEqual(len(RomTestIndexMultiCol.query.filter(attr2=['1', '2']).execute()), 1)
        self.assertEqual(len(RomTestIndexMultiCol.query.filter(attr1='548ef7ee7b77b93ab41ksjh3', attr2='2').execute()), 1)
        self.assertEqual(len(RomTestIndexMultiCol.query.filter(attr1='548ef7ee7b77b93ab41ksjh3', attr2=['2', '1']).execute()), 1)
        self.assertEqual(len(RomTestIndexMultiCol.query.filter(attr1='548ef7ee7b77b93ab41ksjh3', attr2=['1']).execute()), 0)

    def test_namespace(self):
        _ex = {'prefix':True, 'suffix':True, 'keygen':FULL_TEXT}
        class TestNamespace(Model):
            _namespace = 'RomTestNamespace'
            test_i = Integer(index=True)
            test_s = string(unique=True, **_ex)
            test_t = string(index=True, keygen=FULL_TEXT)

        a = TestNamespace(test_i=4, test_s='hello', test_t='this is a test')
        a.save()

        # make sure no strange keys make it through
        self.assertEqual(util.CONNECTION.keys('TestNamespace*'), [])
        # make sure that there are keys named as we wanted them named
        self.assertTrue(len(util.CONNECTION.keys('RomTestNamespace*')) > 0)

        # verify that our indexes work the way we want them to...
        self.assertEqual(a.get_by(test_i=4), [a])
        self.assertEqual(a.get_by(test_i=(3, 5)), [a])
        self.assertEqual(a.get_by(test_i=6), [])
        self.assertEqual(a.get_by(test_s='hello'), a)
        self.assertEqual(a.query.startswith(test_s='hel').all(), [a])
        self.assertEqual(a.query.startswith(test_s='hel0').all(), [])
        self.assertEqual(a.query.endswith(test_s='lo').all(), [a])
        self.assertEqual(a.query.endswith(test_s='llo').all(), [a])
        self.assertEqual(a.query.endswith(test_s='elo').all(), [])
        self.assertEqual(a.get_by(test_t='this'), [a])
        self.assertEqual(a.get_by(test_t=['test', 'blah']), [a])
        self.assertEqual(a.query.count(), 1)

    def test_order_string_index(self):
        class RomTestOrderString(Model):
            test_s = string(index=True, keygen=SIMPLE)
            test_c = string(index=True, keygen=SIMPLE_CI)

        RomTestOrderString(test_s='Hello', test_c='World 2').save()
        RomTestOrderString(test_s='hello', test_c='world 1').save()
        # test case-sensitive ordering
        r1 = RomTestOrderString.query.order_by('test_s').all()
        self.assertEqual(len(r1), 2)

        if sys.version < '2.7':
            self.assertTrue(r1[0].id < r1[1].id)
        else:
            self.assertLess(r1[0].id, r1[1].id)

        # test case-insensitive ordering
        r2 = RomTestOrderString.query.order_by('test_c').all()
        self.assertEqual(len(r2), 2)
        if sys.version < '2.7':
            self.assertTrue(r2[0].id > r2[1].id)
        else:
            self.assertGreater(r2[0].id, r2[1].id)

    def test_string_in_3x(self):
        if six.PY2:
            return
        by = 'hello'.encode('utf-8')
        class RomTestByteString(Model):
            scol = String(unique=True, index=True, suffix=True, keygen=FULL_TEXT)

        RomTestByteString(scol=by).save()
        self.assertTrue(RomTestByteString.get_by(scol=by))
        self.assertEqual(RomTestByteString.query.filter(scol=by).count(), 1)
        self.assertEqual(RomTestByteString.query.endswith(scol=by[1:]).count(), 1)

    def test_hooks(self):
        class RomTestHooks(Model):
            a = Integer(default=0)
            def _before_insert(self):
                self.a += 1
            def _after_insert(self):
                self.a += 4
            def _before_update(self):
                self.a += 16
            def _after_update(self):
                self.a += 64
            def _before_delete(self):
                self.a += 256
            def _after_delete(self):
                self.a += 1024

        d = RomTestHooks()
        self.assertEqual(d.a, 0)
        self.assertTrue(d._new)
        d.save()
        self.assertFalse(d._new)
        self.assertEqual(d.a, 5)
        d.save() # was modified, thanks to our hooks ;)
        self.assertEqual(d.a, 85)
        d.delete()
        self.assertEqual(d.a, 1365)
        self.assertTrue(d._deleted)

        def raise_exception():
            raise ValueError("What!")

        # Make sure that exceptions in the methods cause the operation to fail
        d = RomTestHooks()
        del d.a
        self.assertRaises(TypeError, d.save)
        self.assertTrue(d._new)
        d._before_insert = raise_exception
        self.assertRaises(ValueError, d.save)
        del d._before_insert
        d.a = 0
        d._after_insert = raise_exception
        self.assertRaises(ValueError, d.save)
        self.assertEqual(d.a, 1)
        del d._after_insert
        d.a = 0

        # test for updates
        self.assertFalse(d._new)
        del d.a
        self.assertRaises(TypeError, d.save)
        d._before_update = raise_exception
        self.assertRaises(ValueError, d.save)
        del d._before_update
        d.a = 0
        d._after_update = raise_exception
        self.assertRaises(ValueError, d.save)
        self.assertEqual(d.a, 16)
        del d._after_update
        d.a = 0

        # and again for deletes
        del d.a
        self.assertRaises(TypeError, d.delete)
        d._before_delete = raise_exception
        self.assertRaises(ValueError, d.delete)
        del d._before_delete
        d.a = 0
        d._after_delete = raise_exception
        self.assertRaises(ValueError, d.delete)
        self.assertEqual(d.a, 256)

    def test_data_race(self):
        class RomTestDataRace(Model):
            col = Integer()
        x = RomTestDataRace(col=5)
        x.save()
        session.rollback()
        y = RomTestDataRace.get(x.id)
        self.assertTrue(y)
        self.assertNotEqual(x, y)
        y.col = 6
        y.save()
        session.rollback()
        x.col = 7
        self.assertRaises(DataRaceError, x.save)
        self.assertRaises(DataRaceError, session.commit)
        x.refresh(force=True)
        x.save()
        y.delete()
        self.assertRaises(EntityDeletedError, x.save)
        x.save(force=True)

    def test_keygen2(self):
        def kg2(attr, data):
            keys = set(FULL_TEXT(data.get('a')) or [])
            keys.update(FULL_TEXT(data.get('b')) or [])
            return keys

        class RomTestKeygen2(Model):
            a = string(index=True, keygen2=kg2)
            b = string()

        RomTestKeygen2(a='hello world', b='how are you').save()
        self.assertEqual(RomTestKeygen2.query.filter(a='hello').filter(a='are').count(), 1)

    def test_multiindex(self):
        def kg(val):
            keys = dict.fromkeys(val.split())
            for k in list(keys):
                if k.isdigit():
                    keys['v:'+k] = int(k)
            return keys

        class RomTestMultiindex(Model):
            a = string(index=True, keygen=kg)

        RomTestMultiindex(a='hello world 123').save()
        self.assertEqual(RomTestMultiindex.query.filter(a='hello').count(), 1)
        self.assertEqual(RomTestMultiindex.query.filter(**{'a:v:123':(120, 125)}).count(), 1)

    ## def test_script_flush(self):
        ## c = connect(None)
        ## script = util._script_load('''
            ## return 1
        ## ''')
        ## self.assertEqual(script(c), 1)
        ## c.execute_command('SCRIPT', 'FLUSH', parse="FLUSH")
        ## self.assertEqual(script(c), 1)

    def test_binary2(self):
        class RomTestBinaryData2(Model):
            value = String()
        self.test_binary(RomTestBinaryData2, ''.join(chr(i) for i in range(256)))

    def test_binary(self, RomTestBinaryData=None, bad=None):
        if not bad:
            bad = base64.b64decode(b'UEsDBBQAAAAIACUdX0djd8CXNQEAAAoCAAAYAAAAeGwvd29ya3NoZQ==')
            if six.PY3:
                bad = bad.decode('latin-1')

        if not RomTestBinaryData:
            class RomTestBinaryData(Model):
                value = String()

        d = RomTestBinaryData(value=bad)
        d.save()

        id = d.id
        del d
        session.rollback()
        d = RomTestBinaryData.get(id)
        ## print(type(dv), type(bad))
        if six.PY3:
            # This right here is ridiculous. The 2.x way was 10x better.
            ## print(binascii.hexlify(d.value))
            ## print(binascii.hexlify(bad.encode('latin-1')))
            self.assertEqual(d.value, bad.encode('latin-1'))
        else:
            ## print(d.value.encode('hex'))
            ## print(bad.encode('hex'))
            self.assertEqual(d.value, bad)

    def test_geo(self):
        conn = connect(None)
        version = list(map(int, conn.info()['redis_version'].split('.')))
        if version < [3, 2]:
            print("Skipping geo tests")
            return

        class RomTestGeo(Model):
            tags = String(index=True, keygen=FULL_TEXT)
            lat = Float()
            lon = Float()
            geo_index = [
                # also test out attrdict in here :)
                GeoIndex('basic', lambda x: {'lat':x.lat, 'lon':x['lon']})
            ]

        a = RomTestGeo(lat=0, lon=0, tags='restaurant')
        a.save()
        self.assertEqual(a.query.filter(tags='restaurant').near('basic', .5, 0, 60, 'km').count(), 1)
        self.assertEqual(a.query.filter(tags='gasoline').near('basic', .5, 0, 60, 'km').count(), 0)
        self.assertEqual(a.query.filter(tags='restaurant').near('basic', 0, .5, 60, 'km').count(), 1)
        self.assertEqual(a.query.filter(tags='gasoline').near('basic', 0, .5, 60, 'km').count(), 0)
        self.assertEqual(a.query.filter(tags='restaurant').near('basic', 1, 0, 60, 'km').count(), 0)
        self.assertEqual(a.query.filter(tags='restaurant').near('basic', 0, 1, 60, 'km').count(), 0)

    def _test_filter_performance(self):
        import time
        class RomTestFilterPerformance(Model):
            id = PrimaryKey(index=True)

        ids = []
        for i in range(10000):
            a = RomTestFilterPerformance()
            a.save()
            ids.append(a.id)
        session.rollback()

        t = time.time()
        for i in range(100):
            RomTestFilterPerformance.query.filter(id=(1, 1)).cached_result(30)
        print("\nelapsed: ", time.time()-t)

    def test_recursive_model(self):
        class RomTestRecursiveModel(Model):
            manager = ManyToOne('RomTestRecursiveModel', on_delete='restrict')
            reports = OneToMany('RomTestRecursiveModel')

        a = RomTestRecursiveModel()
        a.save()
        b = RomTestRecursiveModel(manager=a)
        b.save()
        self.assertEqual(a, b.manager)
        self.assertEqual(a.reports, [b])

    def test_lock(self):
        class RomTestModelLock(Model):
            example = Integer()

        a = RomTestModelLock()
        a.save()
        l = util.EntityLock(a, 1, 5)
        self.assertTrue(l.acquire())

        l2 = util.EntityLock(a, 1, 1)
        self.assertFalse(l2.acquire())
        self.assertTrue(l.refresh())
        self.assertTrue(l.release())

    def test_empty_keygen_result(self):
        class RomTestEmptyKeygen(Model):
            col = Text(required=True, index=True, keygen=FULL_TEXT, prefix=True)

        a = RomTestEmptyKeygen(col=u'& &')
        a.save()
        aid = a.id
        del a
        session.rollback()

        b = RomTestEmptyKeygen.get(aid)
        self.assertTrue(b.col)

    def test_type_check_datetime(self):
        class RomTestCheckedColumns(Model):
            c = SaferDateTime()

        # should be okay
        a = RomTestCheckedColumns(c=datetime(1970, 1, 2))
        a.save()

        # also should work, because we need to round trip Redis
        # Fail on setting attributes
        self.assertRaises(InvalidColumnValue, lambda: setattr(a, 'c', 1))
        self.assertRaises(InvalidColumnValue, lambda: setattr(a, 'c', 'blah'))
        self.assertRaises(InvalidColumnValue, lambda: setattr(a, 'c', '1'))
        self.assertRaises(InvalidColumnValue, lambda: setattr(a, 'c', ()))
        self.assertRaises(InvalidColumnValue, lambda: setattr(a, 'c', []))
        self.assertRaises(InvalidColumnValue, lambda: setattr(a, 'c', {}))
        a.c = datetime(1970, 1, 3)
        a.save()

        # Should fail!
        self.assertRaises(InvalidColumnValue, lambda: RomTestCheckedColumns(c=86400))
        self.assertRaises(InvalidColumnValue, lambda: RomTestCheckedColumns(c='86400'))

    ## def test_slorg_example_102(self):
        ## class RomTestSlorg(Model):
            ## col = Text(required=True, index=True, keygen=FULL_TEXT, prefix=True)

        ## d = RomTestSlorg(col='an apple')
        ## d.save()
        ## self.assertEqual(RomTestSlorg.query.startswith(col='an').startswith(col='a').count(), 1)


    ## def test_luis_106(self):
        ## class RomTestLuis1(Model):
            ## col = String(required=True, unique=True, index=True, keygen=rom.SIMPLE)

        ## self.assertEqual(RomTestLuis1(col='Luis').save(), 2)
        ## self.assertNotEqual(RomTestLuis1.get_by(col='Luis'), None)
        ## self.assertNotEqual(RomTestLuis1.query.filter(col='Luis').all(), [])

        ## class RomTestLuis1(Model):
            ## col = String(required=True, unique=True, index=True, keygen=rom.SIMPLE)

        ## self.assertEqual(RomTestLuis1(col='Luis').save(), 2)
        ## self.assertNotEqual(RomTestLuis1.get_by(col='Luis'), None)
        ## self.assertNotEqual(RomTestLuis1.query.filter(col='Luis').all(), [])



def main():
    global_setup()
    try:
        unittest.main()
    finally:
        global_setup()

if __name__ == '__main__':
    main()



#!/usr/bin/env python

try:
    from setuptools import setup
except:
    from distutils.core import setup

long_description = open('README.rst').read()
version = open('VERSION').read().strip()

setup(
    name='rom',
    version=version,
    description='A Redis object mapper for Python',
    author='Josiah Carlson',
    author_email='josiah.carlson@gmail.com',
    url='https://github.com/josiahcarlson/rom',
    packages=['rom'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    license='GNU LGPL v2.1',
    long_description=long_description,
    requires=['redis', 'six'],
    install_requires=['redis', 'six'],
)

